unit UfrmMain;

{

Implementation of the Visualisation described in:

    Focus+Context Display and Navigation Techniques for Enhancing Radial,
    Space-Filling Hierarchy Visualizations

    John Stasko and Eugene Zhang
    GVU Center and College of Computing
    Georgia Institute of Technology

Author: Thomas Rauscher, uni@sinnfrei.at

}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,Contnrs,
  Dialogs, GR32_Image,GR32,G32_Interface, StdCtrls,GR32_Polygons,GR32_Resamplers,
  ExtCtrls, FileCtrl, JvDriveCtrls, JvExStdCtrls, JvListBox,StrUtils,
  TntStdCtrls, TntExtCtrls, Spin, JvBaseDlg, JvSelectDirectory,
  JvComponent, JvBrowseFolder;

type
  TTreeArc=record
    rb,re,ab,ae :double;
  end;
  TTreeArcPos=record
    cur,start,stop :TTreeArc;
  end;
  TTreeElement=class
  private
    function Getleave(i: integer): TTreeElement;
  public
    text :string;
    path :string;
    size,cnt :int64;
    isDir :boolean;
    color :TColor32;
    dt :TDateTime;
    arcpos :array[0..2] of  TTreeArcPos;
//    norm,zoom :TTreeArc;
//    animb,anime :TTreeArc;
    level,depth :integer;
    parent :TTreeElement;
    leaves :TObjectList;
    constructor Create;
    destructor Destroy; override;
    procedure BuildPath(path :string);
    property leave[i :integer] :TTreeElement read Getleave;
  end;
  TfrmMain = class(TForm)
    TntPanel1: TTntPanel;
    lbCur: TLabel;
    lbTotal: TLabel;
    Button2: TButton;
    Button3: TButton;
    dirBox: TJvDirectoryListBox;
    Timer: TTimer;
    rgMode: TTntRadioGroup;
    spMaxL0: TSpinEdit;
    TntLabel1: TTntLabel;
    spMaxL1: TSpinEdit;
    TntLabel2: TTntLabel;
    rgColor: TTntRadioGroup;
    TntPanel2: TTntPanel;
    pb: TPaintBox32;
    edPath: TTntEdit;
    TntLabel3: TTntLabel;
    TntButton1: TTntButton;
    sd: TJvBrowseForFolderDialog;
    gbColorLegend: TTntGroupBox;
    TntShape1: TTntShape;
    TntLabel4: TTntLabel;
    TntShape2: TTntShape;
    TntLabel5: TTntLabel;
    TntShape3: TTntShape;
    TntLabel6: TTntLabel;
    TntShape4: TTntShape;
    TntLabel7: TTntLabel;
    TntShape5: TTntShape;
    TntLabel8: TTntLabel;
    TntShape6: TTntShape;
    TntLabel9: TTntLabel;
    TntShape7: TTntShape;
    TntLabel10: TTntLabel;
    procedure FormCreate(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure pbMouseMove(Sender: TObject; Shift: TShiftState; X,Y: Integer);
    procedure pbClick(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure TimerTimer(Sender: TObject);
    procedure dirBoxClick(Sender: TObject);
    procedure rgModeClick(Sender: TObject);
    procedure rgColorClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure TntButton1Click(Sender: TObject);
  private
    colormode:integer;
    TweenAfirst :array[0..2] of boolean;
    TweenSteps :array[0..2] of integer;
    procedure clearTree(te: TTreeElement; level: integer; id :integer);
    procedure CalcTree(te: TTreeElement; center: TPoint; r_start, r_width, a_start, a_width: double; level: integer;id :integer;setStart :boolean;var depth :integer);
    procedure drawTree(te: TTreeElement; center: TPoint; level: integer; id :integer;tween :double=1.0;aFirst :boolean=false);
    procedure ActivateTree(aNextTree: TTreeElement);
    procedure SetupParam(setcenter :boolean=true);
    procedure colorTree(te: TTreeElement);
  public
     buf :array[0..2] of TBitmap32Ex;
     root :TTreeElement;
     Center :array[0..2] of TPoint;
     Centerb :array[0..2] of TPoint;
     Centere :array[0..2] of TPoint;
     rs,rl,maxlevel :array[0..2] of integer;
     curTree,activeTree,nextTree :TTreeElement;
     openA :double;
  end;

var
  frmMain: TfrmMain;

implementation

uses Math, Types;

{$R *.dfm}

function  ComTree(Item1, Item2: Pointer): Integer;
// compare two tree elements to sort it with directories first, than alphabetecaly

var t1,t2 :TTreeElement;

begin
  t1:=Item1;
  t2:=Item2;
  if t1.isDir=false then begin
    if t2.isDir=true then begin
      result:=1;
    end else begin
      result:=AnsiCompareText(t1.text,t2.text);
    end;
  end else begin
    if t2.isDir=false then begin
      result:=-1;
    end else begin
      result:=AnsiCompareText(t1.text,t2.text);
    end;
  end;
end;

var w :double;

procedure TfrmMain.FormCreate(Sender: TObject);

var r :TKernelResampler;

begin
  w:=0;
  buf[0]:=TBitmap32Ex.Create;
  buf[0].SetSize(pb.width,pb.height);
  buf[1]:=TBitmap32Ex.Create;
  buf[1].SetSize(pb.width,pb.height);
  r:=TKernelResampler.Create(buf[0]);
  r.Kernel:=TMitchellKernel.Create;
  root:=TTreeElement.Create;
end;

{ TTreeElement }
// One Element in the directory Tree

procedure TTreeElement.BuildPath(path: string);
// Build the tree recursively from the current path

var F :TSearchRec;
    r :integer;
    t :TTreeElement;

begin
  size:=0;
  leaves.Clear;
  self.path:=path;
  self.isDir:=true;
  r:=FindFirst(path + '\*.*', faAnyFile ,f);
  while r=0 do begin
    if (F.Name<>'.') and (F.Name<>'..') then begin
      inc(cnt);
      t:=TTreeElement.Create;
      t.text:=F.Name;
      t.parent:=self;
      t.color:=Color32(trunc(Random(255)),trunc(Random(255)),trunc(Random(255)));
      t.dt:=FileDateToDateTime(F.Time);
      if F.Attr=faDirectory then begin
        t.BuildPath(path + '\' + f.Name);
        size:=size + t.Size;
        cnt:=cnt + t.cnt;
      end else begin
        t.size:=F.Size;
        size:=size + F.Size;
      end;
      leaves.add(t);
    end;
    r:=FindNext(f);
  end;
  FindClose(f);
  leaves.Sort(ComTree);
end;

constructor TTreeElement.Create;

begin
  leaves:=TObjectList.Create;
  isDir:=false;
  size:=0;
end;

destructor TTreeElement.Destroy;
begin
  FreeAndNil(leaves);
  inherited;
end;

procedure TfrmMain.CalcTree(te :TTreeElement;center:TPoint;r_start,r_width,a_start,a_width:double;level :integer;id :integer;setStart :boolean; var depth :integer);

// Calculate the positions of each tree element

var  t :TTreeElement;
     i,j :integer;
     hs :string;
     a,b,tw :Double;
     cur :TTreeArc;

begin
  a:=a_start;
  b:=a_start;
  depth:=1;

  cur.rb:=r_start + r_width*(level);
  cur.re:=r_start + r_width*(level+1);
  cur.ab:=a_start;
  cur.ae:=a_start + a_width;

  if setStart then begin
    te.arcpos[id].start.ab:=cur.ab;
    te.arcpos[id].start.ae:=cur.ae;
  end else begin
    te.arcpos[id].cur:=cur;
    te.arcpos[id].stop:=cur;
  end;
  if id=0 then begin
    te.level:=level;
  end;


  for i:=0 to te.leaves.Count-1 do begin
    t:=te.leaves.items[i] as TTreeElement;

    if te.size>0 then begin
      b:=a + a_width * t.size / te.size;
    end else begin
      b:=a;
    end;
    calcTree(t,center,r_start,r_width, a, b-a,level + 1,id,setStart,t.depth);
    if t.depth>=depth then depth:=t.depth+1;
    a:=b;
  end;
end;

procedure TfrmMain.drawTree(te :TTreeElement;center:TPoint;level :integer;id :integer;tween :double=1.0;aFirst :boolean=false);

// Draw the tree elements

var  t :TTreeElement;
     i,j :integer;
//     hs :string;
     cur,bp,ep :TTreeArc;
     tw :double;
     dm :integer;
//  Cache to sum up smaller tree elements
     cache :boolean;
     cacheb :double;

begin
  cache:=false;
  cacheb:=0;
  if maxlevel[id]<=level then exit;
  for i:=0 to te.leaves.Count-1 do begin
    t:=te.leaves.items[i] as TTreeElement;

    if t.isDir then begin
      drawTree(t,center,level + 1,id,tween,aFirst);
    end;
    if tween<1.0 then begin
      if tween<=0.5 then begin
        tw:=tween*2;
      end else begin
        tw:=tween*2 - 1.0;
      end;
      bp:=t.arcpos[id].start;
      ep:=t.arcpos[id].stop;
      if (tween<=0.5) then begin
        if aFirst then begin
          cur.rb:=bp.rb;
          cur.re:=bp.re;
          cur.ab:=bp.ab + tw*(ep.ab - bp.ab);
          cur.ae:=bp.ae + tw*(ep.ae - bp.ae);
        end else begin
          cur.rb:=bp.rb + tw*(ep.rb - bp.rb);
          cur.re:=bp.re + tw*(ep.re - bp.re);
          cur.ab:=bp.ab;
          cur.ae:=bp.ae;
        end;
      end else begin
        if aFirst then begin
          cur.rb:=bp.rb + tw*(ep.rb - bp.rb);
          cur.re:=bp.re + tw*(ep.re - bp.re);
          cur.ab:=ep.ab;
          cur.ae:=ep.ae;
        end else begin
          cur.rb:=ep.rb;
          cur.re:=ep.re;
          cur.ab:=bp.ab + tw*(ep.ab - bp.ab);
          cur.ae:=bp.ae + tw*(ep.ae - bp.ae);
        end;
      end;
      t.arcpos[id].cur:=cur;
    end else begin
      cur:=t.arcpos[id].cur;
    end;
    dm:=4;
    if tween=1.0 then dm:=dm+3;
    if cur.ae-cur.ab>(0.1*2*PI/360) then begin
      if cache then begin
        if cur.rb=0 then begin
          if cur.re>0 then buf[id].Pie(FixedPoint(center),Fixed(cur.re),cacheb,cur.ab,clGray32,dm);
        end else begin
          buf[id].Pie2(FixedPoint(center),Fixed(cur.rb),Fixed(cur.re),cacheb,cur.ab,clGray32,dm);
        end;
        cache:=false;
      end;
      if cur.rb=0 then begin
        if cur.re>0 then buf[id].Pie(FixedPoint(center),Fixed(cur.re),cur.ab,cur.ae,t.color,dm);
      end else begin
        buf[id].Pie2(FixedPoint(center),Fixed(cur.rb),Fixed(cur.re),cur.ab,cur.ae,t.color,dm);
      end;
    end else begin
      if (not cache) and (cur.ae<>cur.ab) then begin
        cacheb:=cur.ab;
        cache:=true;
      end;
    end;
  end;
  if cache then begin
    if cur.rb=0 then begin
      if cur.re>0 then buf[id].Pie(FixedPoint(center),Fixed(cur.re),cacheb,cur.ae,clGray32,dm);
    end else begin
      buf[id].Pie2(FixedPoint(center),Fixed(cur.rb),Fixed(cur.re),cacheb,cur.ae,clGray32,dm);
    end;
    cache:=false;
  end;
end;


procedure TfrmMain.Button2Click(Sender: TObject);

var  i :integer;
     hs :string;
     rw :integer;

begin
  openA:=PI;
  Center[0]:=Point(pb.Width div 2,pb.Height div 2);
  Centerb[0]:=Point(pb.Width div 2,pb.Height div 2);
  Centere[0]:=Point(pb.Width div 2,pb.Height div 2);
  Center[1]:=Point(pb.Width div 2,pb.Height div 2);
  Centerb[1]:=Point(pb.Width div 2,pb.Height div 2);
  Centere[1]:=Point(pb.Width div 2,pb.Height div 2);
  root.free;
  root:=TTreeElement.Create;
  root.BuildPath(edPath.Text);
  lbTotal.Caption:='Size: ' + IntToStr(root.size) + ' Elements: ' + IntToStr(root.cnt);
  root.depth:=1;
  maxlevel[0]:=4;
  maxlevel[1]:=20;
  calcTree(root,Center[0],1,1,0,2*PI,0,0,false,root.depth);
  rgModeClick(self);
end;


function FindTreeElement(r,a :double;start :TTreeElement;id :integer) :TTreeElement;
// Find one specific tree element from the angle and the radius

var i :integer;
    cur :TTreeArc;
    ca :double;

begin
  result:=nil;
  cur:=start.arcpos[id].cur;
  ca:=a - 2*PI;
  if ((cur.ab<=a) and (cur.ae>=a)) or ((cur.ab<=ca) and (cur.ae>=ca)) then begin
    if (cur.rb<=r) then begin
      if (cur.re>=r) then begin
        result:=start;
        exit;
      end else begin
        for i:=0 to start.leaves.Count-1 do begin
          result:=FindTreeElement(r,a,start.leave[i],id);
          if result<>NIL then exit;
        end;
      end;
    end;
  end;
end;


procedure TfrmMain.pbMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);

var r,a,dx,dy :double;
    te :TTreeElement;
    text :string;

begin
  te:=NIL;
  if (activeTree<>NIL) then begin
    dx:=(x-Center[1].X);
    dy:=(y-Center[1].y);
    r:=sqrt(dx*dx + dy*dy);
    a:=-ArcTan2(dy,dx);
    if a<0 then a:=a + 2*PI;
    te:=FindTreeElement(r,a,activeTree,1);
    if (te<>NIL) and (te.level>=maxlevel[1]) then begin
      te:=NIL;
    end;
  end;
  if (te=nil) then begin
    dx:=(x-Center[0].X);
    dy:=(y-Center[0].y);
    r:=sqrt(dx*dx + dy*dy);
    a:=-ArcTan2(dy,dx);
    if a<0 then a:=a + 2*PI;
    te:=FindTreeElement(r,a,root,0);
    if (te<>NIL) and (te.level>=maxlevel[0]) then begin
      te:=NIL;
    end;
  end;
  if (te<>NIL) then begin
    if te.parent<>NIL then begin
      text:=te.parent.path + '\' + te.text;
    end else begin
      text:=te.path;
    end;
  end else begin
    text:='-';
  end;
  curTree:=te;
  lbCur.Caption:=text;
  if pb.hint<>text then begin
    pb.hint:=text;
    Application.CancelHint;
  end;
//  lbV.Caption:=FloatToStr(r) + '  ' + FloatToStr(a);
end;

function TTreeElement.Getleave(i: integer): TTreeElement;
begin
  result:=leaves.items[i] as TTreeElement;
end;


procedure TfrmMain.ActivateTree(aNextTree :TTreeElement);
// Start the animation to put another tree element into the focus

var i :integer;
    ll,cl,dp,rw,start,crs :integer;
    d,ca :double;
    afirst,isParent,frepaint :boolean;
    ht :TTreeElement;
    asteps :integer;

begin
  frepaint:=false;
  nexttree:=aNextTree;
  asteps:=10;
  if nextTree=nil then exit;
  if activeTree=nextTree then exit;
  if activeTree=nil then begin
    ll:=0;
    SetupParam(false);
    Tweensteps[0]:=0;
    frepaint:=true;
  end else begin
    ll:=activeTree.level;
  end;
  cl:=nextTree.Level;
  start:=0;
  afirst:=ll>cl;
  isParent:=false;
  if cl>ll then begin
    ht:=nextTree;
    while ht<>nil do begin
      if ht=activeTree then isParent:=true;
      ht:=ht.parent;
    end;
  end;
  if cl<ll then begin
    ht:=activeTree;
    while ht<>nil do begin
      if ht=nextTree then isParent:=true;
      ht:=ht.parent;
    end;
  end;
  Center[1]:=Point(pb.Width div 2,pb.Height div 2);
  clearTree(root,0,1);
//  nextTree.arcpos[1].cur.rb:=50-30;
//  nextTree.arcpos[1].cur.re:=50;
  nextTree.arcpos[1].cur.rb:=0;
  nextTree.arcpos[1].cur.re:=0;
  ca:=(nextTree.arcpos[0].cur.ab + nextTree.arcpos[0].cur.ae) / 2;
  nextTree.arcpos[1].cur.ab:=ca - openA/2;
  nextTree.arcpos[1].cur.ae:=ca + openA/2;
  dp:=0;
  rw:=(rl[1] - rs[1]) div min(nextTree.depth,maxlevel[1] - nextTree.level);
//        rw:=(rl[i] - rs[i]) div min(activeTree.depth,(maxlevel[i]-activeTree.level));
  crs:=rs[1];
  if rw>30 then begin
    rw:=30;
    if rgMode.ItemIndex=2 then begin
      crs:=rl[1] - rw * min(nextTree.depth,maxlevel[1] - nextTree.level);
    end;  
  end;

  if nextTree.isDir then begin
    dirBox.Directory:=nextTree.path;
  end;
  if not isParent then begin
//    start:=asteps;
    start:=0;
    calcTree(nextTree,Center[1],20,rw,nextTree.arcpos[0].cur.ab,(nextTree.arcpos[0].cur.ae-nextTree.arcpos[0].cur.ab),0,1,true,dp);
    afirst:=false;
  end;
  activeTree:=nextTree;
  Tweensteps[1]:=start;
  TweenAfirst[1]:=afirst;

  calcTree(nextTree,Center[1],crs,rw,nextTree.arcpos[1].cur.ab,openA,0,1,false,dp);
  nextTree.arcpos[1].stop:=nextTree.arcpos[1].cur;

//  nextTree.arcpos[1].stop.rb:=0;
//  nextTree.arcpos[1].stop.re:=0;

  if rgMode.ItemIndex=0 then begin

//    SetupParam;
    Centere[0]:=Point(trunc(pb.Width div 2 - 150*cos(ca)),trunc(pb.Height div 2 +  150*sin(ca)));
    Centere[1]:=Centere[0];
    frepaint:=true;
  end;
  if frepaint then begin
    clearTree(root,0,0);
    rw:=(rl[0] - rs[0]) div min(root.depth,maxlevel[0]);
    calcTree(root,Center[0],rs[0],rw,0,2*PI,0,0,false,root.depth);

    Tweensteps[0]:=start;
    TweenAfirst[0]:=afirst;
  end;

end;

procedure TfrmMain.clearTree(te: TTreeElement; level: integer; id :integer);

// Reset the position of the Tree

var  t :TTreeElement;
     i,j :integer;
     cur :TTreeArc;

begin
  cur.rb:=0;
  cur.re:=0;
  cur.ab:=0;
  cur.ae:=0;
  for i:=0 to te.leaves.Count-1 do begin
    t:=te.leave[i];
    if t.isDir then begin
      clearTree(t,level + 1,id);
    end;
    t.arcpos[id].start:=t.arcpos[id].cur;
    t.arcpos[id].cur.rb:=0;
    t.arcpos[id].cur.re:=0;
    t.arcpos[id].stop.rb:=0;
    t.arcpos[id].stop.re:=0;
  end;
end;

procedure TfrmMain.colorTree(te: TTreeElement);

// Apply colors to the tree

var  t :TTreeElement;
     i,j,tog :integer;
     cur :TTreeArc;
     ext :String;
     h,s,l :single;

begin
  tog:=0;
  for i:=0 to te.leaves.Count-1 do begin
    t:=te.leave[i];
    if t.isDir then begin
      colorTree(t);
    end;
    if colormode=0 then begin
      t.color:=Color32(trunc(Random(255)),trunc(Random(255)),trunc(Random(255)));
    end;
    if colormode=1 then begin
      t.color:=clOlive32;
      if  t.isDir then begin
        t.color:=clLightGray32;
      end;
      ext:=lowercase(ExtractFileExt(t.text));
      if (ext='.jpg') or (ext='.tif') or (ext='.bmp') or (ext='.psd') or (ext='.gif') then begin
        t.color:=clBlue32;
      end;
      if (ext='.avi') or (ext='.mov') or (ext='.divx') or (ext='.xvid') or (ext='.mpg') then begin
        t.color:=clPurple32;
      end;
      if (ext='.dat') or (ext='.img') then begin
        t.color:=clMaroon32;
      end;
      if (ext='.exe') or (ext='.com') or (ext='.cmd') then begin
        t.color:=clRed32;
      end;
      if (ext='.pas') or (ext='.dfm') or (ext='.c')  or (ext='.h') then begin
        t.color:=clYellow32;
      end;
      inc(tog);
      if (tog mod 2)=0 then begin
        RGBtoHSL(t.color,h,s,l);
        l:=2*l / 3;
        t.color:=HSLtoRGB(H, S, L);
      end;
    end;
    if colormode=2 then begin
//      RGBtoHSL(clRed32,H,S,L);
//      RGBtoHSL(clBlue32,H,S,L);
      h:=(now - t.dt) / 365;
      if h>1 then h:=1;
      h:=((1-h) + 2) / 3;
//      RGBtoHSL(Color32(trunc(h*255),0,trunc((1-h)*255)),H,S,L);
      inc(tog);
      s:=1;
      L:=0.5;
      if (tog mod 2)=0 then begin
        l:=2*l / 3;
      end;
      t.color:=HSLtoRGB(H, S, L);
    end;
  end;
end;


procedure TfrmMain.pbClick(Sender: TObject);
begin
  ActivateTree(curTree);
end;

procedure TfrmMain.Button3Click(Sender: TObject);
begin
  if activeTree<>NIL then begin
    if (activeTree.parent<>NIL) and (activeTree.parent<>root) then begin
      ActivateTree(activeTree.parent);
    end else begin
      activeTree:=NIL;
      nextTree:=NIL;
      rgModeClick(sender);
    end;
  end;
end;

procedure TfrmMain.TimerTimer(Sender: TObject);

// The timer is used to draw the animation steps

var i,d,x,y :integer;
    dirty : boolean;
    r,w :double;

begin
  try
    Timer.Enabled:=false;
    dirty:=false;
    for i:=0 to 2 do begin
      if (Tweensteps[i]>=0) and (Tweensteps[i]<=15) then begin
        d:=Tweensteps[i];
        if d>=15 then begin
          Center[i]:=centere[i];
          Centerb[i]:=centere[i];
        end else begin
          Center[i].X:=centerb[i].X + d*(centere[i].X - centerb[i].X) div 15;
          Center[i].Y:=centerb[i].Y + d*(centere[i].Y - centerb[i].Y) div 15;
        end;
      end else begin
        Center[i]:=centerb[i];
      end;
      if Tweensteps[i]<21 then begin
        inc(Tweensteps[i]);
        if (Center[i].X>1) then begin
          buf[i].Clear(Color32(0,0,0,0));
        end;
        if (Tweensteps[i]<21) then begin
          drawTree(root,Center[i],1,i,Tweensteps[i]/(20 + 0.001),TweenAfirst[i]);
        end else begin
          drawTree(root,Center[i],1,i);
        end;
        dirty:=true;
      end;
    end;
    if dirty then begin
      pb.Buffer.Assign(buf[0]);
      pb.Buffer.DrawMode:=dmBlend;
      if activeTree<>NIL then begin
        buf[1].DrawMode:=dmTransparent;
        w:=activeTree.arcpos[0].cur.ab + (activeTree.arcpos[0].cur.ae-activeTree.arcpos[0].cur.ab) / 2;
        r:=activeTree.arcpos[0].cur.rb + (activeTree.arcpos[0].cur.re-activeTree.arcpos[0].cur.rb) / 2;
        x:=center[0].x + trunc(r*cos(w));
        y:=center[0].y - trunc(r*sin(w));
        if activeTree.level<Maxlevel[0] then begin
          buf[1].DrawMode:=dmBlend;
          buf[1].Ellipse(FixedRect(Rect(x-3,y-3,x+3,y+3)),clRed32 );
        end;
        buf[1].DrawTo(pb.Buffer);
      end;
      pb.Invalidate;
    end;
  finally
    Timer.Enabled:=true;
  end;
end;

procedure TfrmMain.dirBoxClick(Sender: TObject);

var hs,hs2 :string;
    i,p,p1,p2 :integer;
    te :TTreeElement;

begin
  te:=root;
  p1:=0;
  hs:=dirBox.Directory;
  p1:=posex('\',hs,p1);
  p2:=posex('\',hs,p1+1)-1;
  if p2<0 then p2:=10000;
  hs2:=copy(hs,1,p2);
  i:=te.leaves.Count-1;
  while (i>=0) do begin
    if te.leave[i].path=hs2 then begin
      te:=te.leave[i];
      i:=te.leaves.Count;
      p2:=posex('\',hs,p2+2)-1;
      if p2<0 then p2:=10000;
      hs2:=copy(hs,1,p2);
    end;
    dec(i);
  end;
  ActivateTree(te);
end;

function VarToInt(v :Variant):longint;

var i:integer;

begin
  i:=0;
 try
  if VarIsNull(v) then result:=0 else begin
   if VarType(v)=varString then Val(v,result,i)
    else result:=v;
  end;
  if i<>0 then result:=0;
 except
  result:=0;
 end;
end;

procedure TfrmMain.SetupParam(setcenter :boolean=true);

// Set the layout for the different modes 

var rw :integer;
    i,dummy,minr :integer;
    ma :double;

begin
  maxlevel[0]:=VarToInt(spMaxL0.Value)+1;
  maxlevel[1]:=VarToInt(spMaxL1.Value)+1;
  minr:=min(pb.Width div 2,pb.height div 2);
  if (activeTree<>NIL) then begin
    ma:=(activeTree.arcpos[0].cur.ab + activeTree.arcpos[0].cur.ae) / 2;
  end else begin
    ma:=0;
  end;
  if setcenter then begin
    Centerb[0]:=Center[0];
    Centerb[1]:=Center[1];
  end;
  if (activeTree<>NIL) or (nextTree<>NIL) then begin
    if rgMode.ItemIndex=0 then begin
      rl[0]:=100;
      rs[0]:=20;
      rl[1]:=300;
      rs[1]:=120;
      if setcenter then begin
        Centere[0]:=Point(trunc(pb.Width div 2 - 150*cos(ma)),trunc(pb.Height div 2 +  150*sin(ma)));
        Centere[1]:=Centere[0];
      end;
      openA:=PI/2;
    end;
    if rgMode.ItemIndex=1 then begin
      rl[0]:=80;
      rs[0]:=30;
      rl[1]:=minr;
      rs[1]:=90;
      if setcenter then begin
        Centere[0]:=Point(pb.Width div 2,pb.Height div 2);
        Centere[1]:=Point(pb.Width div 2,pb.Height div 2);
      end;
      openA:=2*PI;
    end;
    if rgMode.ItemIndex=2 then begin
      rl[0]:=minr;
      rs[0]:=150;
      rl[1]:=150;
      rs[1]:=20;
      if setcenter then begin
        Centere[0]:=Point(pb.Width div 2,pb.Height div 2);
        Centere[1]:=Point(pb.Width div 2,pb.Height div 2);
      end;
      openA:=2*PI;
    end;
  end else begin
      rl[0]:=minr;
      rs[0]:=50;
      rl[1]:=0;
      rs[1]:=0;
      if setcenter then begin
        Centere[0]:=Point(pb.Width div 2,pb.Height div 2);
        Centere[1]:=Point(pb.Width div 2,pb.Height div 2);
      end;  
      openA:=2*PI;
  end;
end;

procedure TfrmMain.rgModeClick(Sender: TObject);

var rw :integer;
    i,dummy :integer;
    ma :double;

begin
  SetupParam;
  for i:=0 to 1 do begin
    clearTree(root,0,i);
    if i=0 then begin
      rw:=(rl[i] - rs[i]) div min(root.depth,maxlevel[i]);
      calcTree(root,Center[i],rs[i],rw,0,2*PI,0,i,false,root.depth);
      Tweensteps[i]:=0;
      TweenAfirst[i]:=false;
    end;
    if (i=1) and (activeTree<>NIL) then begin
      rw:=(rl[i] - rs[i]) div max(min(activeTree.depth,(maxlevel[i]-activeTree.level)),1);
      ma:=(activeTree.arcpos[0].cur.ab + activeTree.arcpos[0].cur.ae) / 2;
      calcTree(activeTree,Center[i],rs[i],rw,ma - openA/2,openA,0,i,false,dummy);
      Tweensteps[i]:=0;
      TweenAfirst[i]:=false;
    end;
  end;
end;

procedure TfrmMain.rgColorClick(Sender: TObject);
begin
  colormode:=rgColor.ItemIndex;
  TweenSteps[0]:=20;
  TweenSteps[1]:=20;
  colorTree(root);
  gbColorLegend.Visible:=rgColor.ItemIndex=1;
end;

procedure TfrmMain.FormResize(Sender: TObject);
begin
  buf[0].SetSize(pb.width,pb.height);
  buf[1].SetSize(pb.width,pb.height);
end;

procedure TfrmMain.TntButton1Click(Sender: TObject);
begin
  if sd.Execute then begin
    edPath.Text:=sd.Directory;
    if copy(edPAth.Text,length(edPAth.Text),1)='\' then begin
      edPAth.Text:=copy(edPAth.Text,1,length(edPAth.Text)-1);
    end;
  end;
end;

end.
